import re
from torch import rand, einsum, allclose, manual_seed, Tensor
from opt_einsum import contract, contract_expression
from typing import List
from timeit import repeat


manual_seed(0)

B = rand(10, 5)
D = rand(5, 20)
E = rand(20, 30)

C = einsum("ij,jk->ik", D, E)
A_naive = einsum("ij,jk->ik", B, C)

A_expanded = einsum("ij,jk,kl->il", B, D, E)

print(allclose(A_naive, A_expanded))

def is_tensor(x):
    return isinstance(x, Tensor)


def is_nested(x):
    return isinstance(x, (tuple, list))


def is_equation(x):
    return isinstance(x, str)


def einsum_nested_naive(expression, optimize=False) -> Tensor:
    """Naively evaluate a nested einsum expression.

    Evaluate unnested expressions until no sub-expressions remains.

    Args:
        expression: List describing a (potentially nested) einsum expression.
            The head is an equation, and the tail consists of tensors or more
            einsum expressions.
        optimize: If ``True``, use ``opt_einsum.contract`` to evaluate unnested
            expressions. Otherwise, use ``torch.einsum``. Default: ``False``.

    Returns:
        Result tensor.
    """
    equation, operands = expression[0], expression[1:]

    if not is_equation(equation):
        raise ValueError(f"Invalid equation as first entry: {equation}.")

    operands_flat = []

    for op in operands:
        if is_tensor(op):
            operands_flat.append(op)
        elif is_nested(op):
            operands_flat.append(einsum_nested_naive(op, optimize=optimize))
        else:
            raise ValueError(f"Expect Tensor or einsum expression, got {op}.")

    contraction_func = contract if optimize else einsum

    return contraction_func(equation, *operands_flat)

A = [
    "ij,jk->ik",
    B,
    ["ij,jk->ik", D, E],
]
A_nested_naive = einsum_nested_naive(A)

print(allclose(A_nested_naive, A_naive))

def equation_to_groups(equation: str) -> List[str]:
    return re.split(",|->", equation)


def groups_to_equation(groups: List[str]) -> str:
    return "->".join([",".join(groups[:-1]), groups[-1]])

equation = "ij,jk,kl->il"
groups = equation_to_groups(equation)

print(f"Groups: {groups}")
print(f"Equation from groups: '{groups_to_equation(groups)}'")

def rename_out_indices(equation, new_name, blocked=None):
    """Rename output indices into new_name, respecting blocked indices."""
    out_indices = equation_to_groups(equation)[-1]

    blocked = set() if blocked is None else blocked.copy()
    blocked = blocked.union(set(new_name)).union(set(out_indices))

    equation = _rename_indices(equation, blocked=blocked)

    out_indices = equation_to_groups(equation)[-1]

    for out_idx, new_idx in zip(out_indices, new_name):
        equation = equation.replace(out_idx, new_idx)

    return equation


def _rename_indices(equation, blocked=None):
    """Rename indices of an equation if they are blocked."""
    groups = equation_to_groups(equation)
    indices = set("".join(groups))

    blocked = set() if blocked is None else blocked.copy()
    blocked = blocked.union(indices)

    rename_indices = indices.intersection(blocked)

    for idx, new_name in zip(rename_indices, _get_free_characters(blocked=blocked)):
        equation = equation.replace(idx, new_name)

    return equation


def _get_free_characters(blocked=None):
    """Yield characters that are not in blocked, starting with a."""
    blocked = set() if blocked is None else blocked

    first = ord("a")
    shift = 0

    while True:
        char = chr(first + shift)
        if char not in blocked:
            yield char
        shift += 1

print(rename_out_indices("ij,jk->ik", "jk"))

print(rename_out_indices("ij,jk->ik", "jk", blocked={"c"}))

def insert_at(expression, pos):
    """In-place insert the sub-expression at the given position."""
    equation, operands = expression[0], expression[1:]
    groups = equation_to_groups(equation)
    blocked = set("".join(groups))

    inner_equation, inner_operands = operands[pos][0], list(operands[pos][1:])
    inner_equation = rename_out_indices(inner_equation, groups[pos], blocked=blocked)
    inner_groups = equation_to_groups(inner_equation)

    out_operands = operands[:pos] + inner_operands + operands[pos + 1 :]
    out_groups = groups[:pos] + inner_groups[:-1] + groups[pos + 1 :]
    out_equation = groups_to_equation(out_groups)

    return [out_equation] + out_operands

# strings instead of tensors for more readable printing
expression = ["ij,jk->ik", "B", ["ij,jk->ik", "D", "E"]]

print(insert_at(expression, 1))

# strings instead of tensors for more readable printing
expression = ["ij,jk->ik", "B", ["ij,jk->ik", ["ab,bk->ak", "D", "E"], "F"]]

print(insert_at(expression, 1))
print(insert_at(insert_at(expression, 1), 1))

def einsum_expand(expression):
    """Expand a nested list of ``einsum`` expressions into a single one.

    An ``einsum`` expression is defined via a tuple/list ``(equation, *operands)``
    where operands are either ``Tensor``s or ``einsum`` expressions.

    Args:
        expression: List describing a (potentially nested) einsum expression.
            The head is an equation, and the tail consists of tensors or more
            einsum expressions.

    Returns:
        An unnested expression consisting of an equation and its operands.
    """
    expression = list(expression)

    def expandable_pos(expression):
        for pos, subexpression in enumerate(expression[1:]):
            if is_nested(subexpression):
                return pos
        raise ValueError("No expandable expression found.")

    while any(is_nested(subexpression) for subexpression in expression[1:]):
        pos = expandable_pos(expression)
        expression = insert_at(expression, pos)

    return expression

# einsum_expand handles list- and tuple-valued expressions
A = ("ij,jk->ik", B, ["ij,jk->ik", D, E])

A_naive = einsum_nested_naive(A)
A_expanded = einsum(*einsum_expand(A))

print(allclose(A_naive, A_expanded))

dim = 30
inner_dim = dim // 10

I = rand(dim, dim, dim, dim)
A = rand(dim, inner_dim)
B = rand(inner_dim, dim)
C = ["ij,jk->ik", A, B]

expression = ["pi,qj,ijkl,rk,sl->pqrs", C, C, I, C, C]

def naive_torch():
    """Naively evaluate with `torch.einsum`."""
    return einsum_nested_naive(expression, optimize=False)


def naive_opt_einsum():
    """Naively evaluate with `opt_einsum.contract`."""
    return einsum_nested_naive(expression, optimize=True)


# NOTE Left out for faster execution
# def expand_torch():
#     """Expand expression then evaluate with `torch.einsum`."""
#     return einsum(*einsum_expand(expression))


def expand_opt_einsum():
    """Expand expression then evaluate with `opt_einsum.contract`."""
    return contract(*einsum_expand(expression))

print(
    allclose(naive_torch(), naive_opt_einsum())
    # NOTE Left out for faster execution
    # and allclose(naive_torch(), expand_torch())
    and allclose(naive_torch(), expand_opt_einsum())
)

REPEAT, NUMBER = 10, 20

naive_torch_time = min(repeat(naive_torch, repeat=REPEAT, number=NUMBER))
# NOTE Left out for faster execution
# expand_torch_time = min(repeat(expand_torch, repeat=REPEAT, number=NUMBER))
naive_opt_einsum_time = min(repeat(naive_opt_einsum, repeat=REPEAT, number=NUMBER))
expand_opt_einsum_time = min(repeat(expand_opt_einsum, repeat=REPEAT, number=NUMBER))

best_overall = min(
    naive_torch_time,
    # NOTE Left out for faster execution
    # expand_torch_time,
    naive_opt_einsum_time,
    expand_opt_einsum_time,
)

print(
    f"Naive  + torch.einsum        : {naive_torch_time:.5f} s "
    + f"(x {naive_torch_time / best_overall:.2f})"
)
# NOTE Left out for faster execution
# print(
#     f"Expand + torch.einsum        : {expand_torch_time:.5f} s "
#     + f"(x {expand_torch_time / best_overall:.2f})"
# )
print(
    f"Naive  + opt_einsum.contract : {naive_opt_einsum_time:.5f} s "
    + f"(x {naive_opt_einsum_time / best_overall:.2f})"
)
print(
    f"Expand + opt_einsum.contract : {expand_opt_einsum_time:.5f} s "
    + f"(x {expand_opt_einsum_time / best_overall:.2f})"
)
